#include <math.h>
#include "box.h"

//*********************************************************************
//******************       CONSTRUCTEURS         **********************
//*********************************************************************
box::box ()
{	dim=1;
	data=new interval[1];
	data[0].isInterval=false;
	isEmpty=true;
    next=NULL;
}
//----------------------------------------------------------------------
box::box (int dim1,bool EstUneBoite=true)
{	dim=dim1;
	data=new interval[dim];
	for (int i=0;i<=dim-1;i++) data[i].isInterval=EstUneBoite;
	isEmpty=false;
    next=NULL;
}
//----------------------------------------------------------------------
box::box (const box &v)
{	if (&v == this) return;
	dim=v.dim;
	isEmpty=v.isEmpty;
	data=new interval[dim];
	for (int k=1; k<=dim; k++) (*this)[k]=v[k];
    //next=NULL;
}
//----------------------------------------------------------------------
box::~box () {delete [] data;}
//----------------------------------------------------------------------
//*********************************************************************
//******************    FUNCTIONS MEMBRES     *************************
//*********************************************************************
interval& box::operator[] (int i)  const
	{		return data[i-1];
	}
//----------------------------------------------------------------------
void box::Resize (int dim1)
{	box a(dim1,true);
	for (int k=1; (k<=dim1)&&(k<=dim); k++)  {a[k]=(*this)[k];}
	delete [] data;
    isEmpty=false;
	data=new interval[dim1];
    for (int k=1; (k<=dim1)&&(k<=dim); k++)  {(*this)[k]=a[k];}
    dim=dim1;
}

//----------------------------------------------------------------------
void box::operator= (box v)
{	if (this == &v) return;
	delete [] data;
	dim=v.dim;
	isEmpty=v.isEmpty;
	data=new interval[dim];
	for (int k=1; k<=dim; k++)
      (*this)[k]=v[k];
}
//----------------------------------------------------------------------
box operator+(const box& v,const box& w)
	{	box r(v.dim);
		for (int k=1; k<=r.dim; k++) r[k]=v[k]+w[k];
		return (r);
	}
//----------------------------------------------------------------------
box operator-(const box& v)
	{
		box r(v.dim);
		for (int k=1; k<=r.dim; k++) r[k]=-v[k];
		return (r);
	}
//----------------------------------------------------------------------
box operator-(const box& v,const box& w) {return (v+(-w)); }
//----------------------------------------------------------------------
box operator*(const interval& a,const box& v)
	{	box r(v.dim);
		for (int k=1; k<=r.dim; k++) r[k]=a*v[k];
		return (r);
	}
//----------------------------------------------------------------------
box	operator*(const box& v,const interval& a)
	{return (a*v);}
//----------------------------------------------------------------------
ostream& operator<<(ostream& os, const box& v)
{   cout << "Box \n";
	cout << "   dim="<<v.dim <<'\n';
    if (v.isEmpty) os<<"EmptyBox";
	for (int i=1; i<=v.dim; i++)
		os << "  " <<i<< ": "<<v[i]<< "\n";
	return (os);
}
//*********************************************************************
//******************    FONCTION NON MEMBRES     **********************
//*********************************************************************
int AxePrincipal(box& x)
	{	double widthmax=Width(x[1]);
		int kmax=1;
		for (int k=2; k<=x.dim; k++)
		{	if (Width(x[k])>widthmax)
			{ kmax=k;widthmax=Width(x[k]);};
		}
		return kmax;
	}
int AxePrincipal(box& x, box& V)
{ int k=1;
  interval I=x[k];
  double widthmax=Width(I)*fabs(Inf(V[k]));
  int kmax=k;
  for (k=2; k<=x.dim; k++)
  {	if (Width(x[k])*fabs(Inf(V[k]))>widthmax)
  	{ kmax=k;widthmax=Width(x[k])*fabs(Inf(V[k]));};
  	//if (w<0) AfxMessageBox ("Erreur axe principal1");
  }
  return kmax;
}
//----------------------------------------------------------------------
void Bisect(box& x, box& x1, box& x2)
{	BisectAlong(x,x1,x2,AxePrincipal(x));	}
//----------------------------------------------------------------------
void BisectRand(box& x, box& x1, box& x2)
{	int i=1+ (rand()%x.dim);
    x1 = x;	x2 = x;
	double mx = Center(x[i]);
	(x1[i]).SetSup(mx);
	(x2[i]).SetInf(mx);
}
//----------------------------------------------------------------------
void Bisect(box& x, box& x1, box& x2,box& V)
{   BisectAlong(x,x1,x2,AxePrincipal(x,V));	}
//----------------------------------------------------------------------
void BisectAlong(box& x, box& x1, box& x2, int i)
{	x1 = x;	x2 = x;
	double mx = Center(x[i]);
	(x1[i]).SetSup(mx);
	(x2[i]).SetInf(mx);
}
//----------------------------------------------------------------------
void BisectHere(box& x, box& x1, box& x2, int i, double here) //Used by the 3-B consistency
{	x1 = x;	x2 = x;
	(x1[i]).SetSup(here);
	(x2[i]).SetInf(here);
}
//----------------------------------------------------------------------
box Center(const box& v)
	{   interval I;
		box r(v.dim);
		for (int k=1; k<=r.dim; k++) {I=v[k];	r[k]=Center(I);}
		return (r);
	}
//----------------------------------------------------------------------
bool Disjoint (const box& a,const box& b)
{   for (int i=1;i<=a.dim;i++)
    { if (Disjoint(a[i],b[i])) return true; }
    return false;
}
//----------------------------------------------------------------------
iboolean In(box& x,box& y)
	{
		if (x.isEmpty) return itrue;
		if (y.isEmpty) return ifalse;
	    iboolean r=itrue;
		for (int k=1; k<=x.dim; k++)
		{	iboolean in1=In(x[k],y[k]);
			if		(in1==false)   return ifalse;
			else if (in1.value==iperhaps) r=iperhaps;
		}
		return r;
	}
//--------------------------------------------------------------------
double Eloignement(box& x,box& y)  //returns a distance between boxes x and y
	{	if ((x.isEmpty)||(y.isEmpty)) return oo;
        double e=0;
	    for (int k=1; k<=x.dim; k++)
		{	e=max(e,Eloignement(x[k],y[k]));}
		return e;
	}
//--------------------------------------------------------------------
double Eloignement2(box& x,box& y)
	{	// prend le point x1 de x qui est le plus eloigne de [y] et renvoie la
        // distance de x1 avec [y]
        if (Subset(x,y)) return 0;
        double e=0;
	    for (int k=1; k<=x.dim; k++)
		{   double e1=0;
            double xinf=Inf(x[k]);
            double xsup=Sup(x[k]);
            double yinf=Inf(y[k]);
            double ysup=Sup(y[k]);
            if (xsup>ysup) e1=max(e1,xsup-ysup);
            if (xinf<yinf) e1=max(e1,yinf-xinf);
            e=max(e,e1);
        }
		return e;
	}
double EloignementRelatif2(box& x,box& y)
	{	// prend le point x1 de x qui est le plus eloigne de [y] et renvoie la
        // distance de x1 avec [y]
        if (Subset(x,y)) return 0;
        if (y.isEmpty) return 99999999;
        double e=0;
	    for (int k=1; k<=x.dim; k++)
		{   double e1=0;
            double xinf=Inf(x[k]);
            double xsup=Sup(x[k]);
            double yinf=Inf(y[k]);
            double ysup=Sup(y[k]);
            if (xsup>ysup) e1=max(e1,fabs((xsup-ysup)/ysup));
            if (xinf<yinf) e1=max(e1,fabs((yinf-xinf)/yinf));
            e=max(e,e1);
        }
		return e;
	}
//----------------------------------------------------------------------
box Inf(box& v)
{   interval I;
	box r(v.dim);
	for (int k=1; k<=r.dim; k++) {I=v[k];	r[k]=Inf(I);}
	return (r);
}
//----------------------------------------------------------------------
box Inter(const box& a,const box& b)
{   box r(a.dim);
    if ((a.isEmpty)||(b.isEmpty))
        {r.isEmpty=true;
        return r;};
    for (int k=1; k<=r.dim; k++)
           {    r[k]=Inter(a[k],b[k]);
                if (r[k].isEmpty) {r.isEmpty=true;return r;}
           }
	return (r);
}
//----------------------------------------------------------------------
void Phi0(box& B0,box* pP) //0-intersection (voir qminimax)
{   if (pP==NULL) {return;}
    B0=Inter(B0,*pP);
    Phi0(B0,pP->next);
}
//----------------------------------------------------------------------
void Phi1(box& B0, box& B1, box* pP) //1-intersection (voir qminimax)
{// Hyp: B1 contient B0
    if (pP==NULL) {return;}
    box B11=Inter(*pP,B1);
    B1=Union(B0,B11);
    B0=Inter(B0,*pP);
    Phi1(B0,B1,pP->next);
}
//----------------------------------------------------------------------
void Phi2(box& B0, box& B1,box& B2,box* pP) //2-intersection
{// Hyp: B2 contient B1 qui contient B0
    if (pP==NULL) {return;}
    box B21=Inter(*pP,B2);   B2=Union(B1,B21);
    box B11=Inter(*pP,B1);   B1=Union(B0,B11);
    B0=Inter(B0,*pP);
    Phi2(B0,B1,B2,pP->next);
}
//----------------------------------------------------------------------
box Inflate(box& a, double eps)
{   box r(a.dim);
    r.isEmpty=a.isEmpty;
    for (int k=1; k<=r.dim; k++)
           {    r[k]=Inflate(a[k],eps);}
	return (r);
}
//----------------------------------------------------------------------
bool Prop(box& x,box& y)  // Normalement x is a subset of y (used in SIVEX)
	{   if (x.isEmpty) return false;
		if (y.isEmpty) return false;
	    for (int k=1; k<=x.dim; k++)
            if (Inf(x[k])==Inf(y[k]) ||  (Sup(x[k])==Sup(y[k])) ) return (true);
        return false;
	}
//----------------------------------------------------------------------
box Sup(box& v)
{   interval I;
	box r(v.dim);
	for (int k=1; k<=r.dim; k++) {I=v[k];	r[k]=Sup(I);}
	return (r);
}
//----------------------------------------------------------------------
box Union(box& a, box& b)
{   box r(a.dim);
    if (a.isEmpty) return (b);
    if (b.isEmpty) return (a);
	for (int k=1; k<=r.dim; k++)
           {    r[k]=Union(a[k],b[k]);
                if (r[k].isEmpty) r.isEmpty=true;}
	return (r);
}
//----------------------------------------------------------------------
bool Subset(box& x,box& y)
{	if (x.isEmpty) return true;
 	if (y.isEmpty) return false;
    bool b=true;
 	for (int k=1; k<=x.dim; k++) b = b && Subset(x[k],y[k]);
 	return (b);
}
//----------------------------------------------------------------------
bool Subset(box& x,box& y,double epsilon)
{	if (Subset(x,y)==false) return false;
    for (int k=1; k<=x.dim; k++)
       if (Subset(x[k],y[k],epsilon)) return true;
    return (false);
}
//----------------------------------------------------------------------
void Sucre(box& P, box& S)    //Sucre le bout de P qui est dans S
{   int sum=0;
    int j=1;
    for (int i=1;i<=P.dim;i++)
         {if (Subset(P[i],S[i])) sum++; else j=i;}
    if (sum==(P.dim)-1)
       if (!Subset(S[j],P[j])) Sucre(P[j],S[j]);
};
//----------------------------------------------------------------------
double Volume(box& x)
{   double vol=1;
    for (int i=1;i<=x.dim;i++) vol=vol*Width(x[i]);
    return vol;
};
//----------------------------------------------------------------------
double Width(box& x)	{return Width(x[AxePrincipal(x)]);	}
//----------------------------------------------------------------------
double Width(box& x,box& V)
	{	int i=AxePrincipal(x,V);
        return (Width(x[i])*fabs(Inf(V[i])));	}
//----------------------------------------------------------------------
interval ProduitScalaire(box& u,box& v)
{  interval somme=0;
   for (int i=1; i<=u.dim; i++)  somme=somme+u[i]*v[i];
   return (somme);
}
//----------------------------------------------------------------------
interval Determinant(box& u,box& v)
{   interval u1=u[1];
    interval v2=v[2];
    interval v1=v[1];
    interval u2=u[2];
    interval r=u1*v2-v1*u2;
    return u1*v2-v1*u2; }
//----------------------------------------------------------------------
// Calcule l'angle entre 2 vecteur de dimension 2
double Angle (box& v, box& w)
// Attention, il faut des vecteurs et non des box et les vecteurs doivent etre
// de dimension 2
{   if ((v.dim!=2)||(w.dim!=2))
    {
       cout<<"error";
    }
    double n2,costeta,sinteta;
    interval nv=Norm(v);
    interval nw=Norm(w);
    interval nvw=nv*nw;
    n2=ToReel(nvw);
    costeta=ToReel(ProduitScalaire(v,w)/n2);
    sinteta=ToReel(Determinant(v,w))/n2;
    return(Arccossin(costeta,sinteta));
}

interval Norm(box& x)
{ int i;
  interval r=0;
  for (i=1; i<=x.dim; i++) r=r+Sqr(x[i]);
  return(Sqrt(r));
}


//----------------------------------------------------------------------
box Zeros(int d)
{   box r(1,d);
	for (int k=1; k<=d; k++) r[k]=0;
	return (r);
}








